#ifndef DICTIONARY_H
#define DICTIONARY_H

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <Arduino.h>


template <typename T>
class Dictionary
{
public:
    ~Dictionary()
    {
        if(_head == this)
            dispose();
        free(_parameter);
        //_parameter = NULL; Useless, just my c habits
        delete _value;
        //_value = NULL; //Useless, just my c habits
    }

    boolean addParameter(const char *parameter, T *value)
    {
        Dictionary *dictionaryNode = new Dictionary(parameter, value);
        return addNewNodeAtTheEnd(dictionaryNode);
    }
    boolean addParameter(const char *parameter, T value)
    {
        Dictionary *dictionaryNode = new Dictionary(parameter, new T(value));
        return addNewNodeAtTheEnd(dictionaryNode);
    }
    boolean deleteParameter(const char *parameter)
    {
        if(_head->_next == NULL) return false;

        Dictionary *cursor = _head, *toDelete(NULL);

        while(!isListEmpty(cursor->_next))
        {
            if(strcmp(cursor->_next->_parameter, parameter) == 0)
            {
                toDelete = cursor->_next;
                cursor->_next = cursor->_next->_next;
                delete toDelete;
                return true;
            }
            cursor = cursor->_next;
        }
        return false;
    }
    Dictionary get(const char *parameter)
    {
        if(isListEmpty(_head->_next))return Dictionary();

        Dictionary *cursor = _head->_next;

        while(!isListEmpty(cursor))
        {
            if(strcmp(cursor->_parameter,parameter) == 0)
                return *cursor;
            cursor = cursor->_next;
        }

        return Dictionary();
    }
    Dictionary operator()(const char *parameter)
    {
        return get(parameter);
    }
    Dictionary get(const unsigned int index)
    {
        unsigned int position(0);
        if(isListEmpty(_head->_next))return Dictionary();

        Dictionary *cursor = _head->_next;

        while(!isListEmpty(cursor))
        {
            if(position++ == index)
                return *cursor;
            cursor = cursor->_next;
        }

        return Dictionary();
    }
    Dictionary operator()(const unsigned int index)
    {
        return get(index);
    }
    unsigned int count()
    {
        unsigned int counter(0);
        if(isListEmpty(_head->_next))return counter;

        Dictionary *cursor = _head->_next;
        while(!isListEmpty(cursor))
        {
            counter++;
            cursor = cursor->_next;
        }
        return counter;
    }
    void clear() {this->dispose();}
    void dispose()
    {
        if(isListEmpty(_head->_next))return;

        Dictionary *cursor = _head->_next, *toDelete(NULL);

        while(!isListEmpty(cursor))
        {
            toDelete = cursor;
            cursor = cursor->_next;

            delete toDelete;
        }
        _head = this;
        _next = NULL;
    }
    const char *stringValue() const {return _value == NULL ? "" : _value->toString();}
    T getValue(){return T(*_value);}
    T* getValueRef(){return _value;}
    const char *getParameter() const{return _parameter == NULL ? "" : _parameter;}
protected:
    Dictionary() :_parameter(NULL), _value(NULL), _next(NULL), _head(this){}
    Dictionary(const char *parameter, T *value) : Dictionary()
    {
        //We copy  the parameter and the value
        _parameter = (char *) malloc((strlen(parameter) * sizeof(char)) + 1); //+1 for the string terminating character

        strcpy(_parameter, parameter);
        _value = value;
    }
    Dictionary(Dictionary const& dictionaryToCopy) //Copy constructor needed because of pointers
    {
        _head = NULL;
        _next = NULL;

        _parameter = (char *) malloc((strlen(dictionaryToCopy._parameter) * sizeof(char)) + 1); //+1 for the string terminating character
        _value = new T(*(dictionaryToCopy._value));

        strcpy(_parameter, dictionaryToCopy._parameter);
    }
    boolean addNewNodeAtTheEnd(Dictionary *node)
    {
        if(node == NULL) return false;

        node->_head = _head; //Every node should point to the first node

        if(_next == NULL) //This is our first node then
        {
            _next = node;
            return true;
        }

        // /!\ We have to work with the _next reference in the loop, if we don't it won't work as expected
        Dictionary *cursor = _head;
        while(!isListEmpty(cursor->_next))
        {
            if(strcmp(cursor->_next->_parameter, node->_parameter) == 0)//If we find the same parameter name, we replace it
            {
                delete (cursor->_next->_value);
                cursor->_next->_value = new T(*(node->_value));
                if(cursor->_next->_value == NULL)return false;

                delete node;
                
                return true;
            }
            cursor = cursor->_next;
        }
        cursor->_next = node;

        return true;
    }
    boolean isListEmpty(Dictionary *node) {return node == NULL;}
    
    char *_parameter;
    T *_value;
    Dictionary *_next;
    Dictionary *_head;
private:
};



#endif //DICTIONARY_H
