#ifndef WEBSERVERMANAGER_H
#define WEBSERVERMANAGER_H

#include <WiFiServer.h>
#include <WiFiClient.h>
#include "SDCardManager.h"
#include "definition.h"

class WEBServerManager
{
  public:
    WEBServerManager(unsigned int port = 80, SDCardManager *sdCardManager = NULL);

    boolean runServer();
    unsigned int getPort() const;
  protected:
  private:
    enum ClientStatus {NOT_HANDLED, HANDLED, NEW, WAITING_FOR_CLIENT, QUERY_PARSED, RESPONSE_SENT};
    enum HttpRequestMethod {UNDEFINED, GET, HEAD, POST, PUT, DELETE, CONNECT, OPTIONS, TRACE, PATCH};
    enum HttpVersion {UNKNOWN, HTTP_0_9, HTTP_1_1, HTTP_1_0, HTTP_2_0};
    enum HttpParserStatus {INIT, LINE_BREAK, HTTP_VERB_SECTION, HTTP_RESOURCE_SECTION, HTTP_VER_SECTION, BODY_SECTION, IGNORED, ERROR};
    enum HttpMIMEType{TEXT_PLAIN, TEXT_CSS, TEXT_HTML, TEXT_JAVASCRIPT};

    struct HttpRequestData{
      HttpRequestMethod HRM;
      HttpVersion HV;
      char *httpResource;
      char *httpBody;
    };
    boolean parseQuery(WiFiClient *wifiClient);
    boolean sendPageToClientFromSdCard(WiFiClient *wifiClient);
    char *addChar(char *pointer, const char character);
    HttpRequestMethod getHttpVerbEnumValue(const char *parseBuffer);
    HttpVersion getHttpVersionEnumValue(const char *parseBuffer);
    char *getFilePathByHttpResource(const char *res);
    char *getFileExtension(char *name);
    HttpMIMEType getMIMETypeByExtension(const char *extension);
    char *getHTTPHeader(HttpMIMEType httpMIMEType, unsigned long size);
    
    WiFiServer _wifiServer;
    WiFiClient _wifiClient;//One client only, maybe replaced with a fifo in the future
    SDCardManager *_sdCardManager;

    HttpRequestData _httpRequestData;
    
    ClientStatus _clientState;
    HttpParserStatus _httpParserState;
    unsigned long _clientTimeout;
    unsigned int _port;
};

#endif //WEBSERVERMANAGER_H
